/*
	csstok.c

	jmd 6/90

	% stock market simulator of the C-scape Demo.

	C-scape 4.0 Example Program
	Copyright (c) 1992 by Liant Software Corporation.
	ALL RIGHTS RESERVED.

	OVERVIEW:
	---------
	This program is a simulation of a simple stock exchange.  The user
	can buy and sell shares of various stocks as their values rise
	and fall.

	The program demonstrates the use of the kb_Idle function
	to update information on the display while the user enters data.
	It also uses auxiliary functions to pass messages between windows (seds).

	The program is structured in the following manner.  There are three
	main objects: the PriceBox, the CashBox, and the BuySellBox.  In addition
	there is a keyboard idle function, idle_Stock, that generates
	events at timed intervals.  Communication between the various objects
	and the idle function is handled by passing messages.  The auxiliary
	functions of the various objects receive the messages and process them.

	The CashBox consists of the ticker tape, time of day clock,
	stock average indicator, and the current cash indicators.

	The PriceBox consists of the list of owned stocks and their
	current prices.  The PriceBox sed allows the user to select stocks
	to buy and sell (by calling the BuySellBox).

	The current stock information is maintained in the list of
	structures, 'stk_list'.  Stocks are referred to by their index
	number in the stk_list.

	The BuySellBox is activated whenever the user wishes to buy or
	sell shares of a stock.  It lets the user enter the number of
	shares to buy or sell, and informs the CashBox about the
	transaction.

	The following messages are used:

		STK_CLOCKTICK		a clock tick occurred, update the ticker tape
							or generate a price change
		STK_PRICECHANGE		a stock price has changed
		STK_BUYSELLREQ		a request to buy or sell shares
		STK_BUYSELL			change the number of owned shares
		STK_UPDATESTATS		update the net worth and stock average indicator
		STK_LASTHOUR		last hour of trading indicator
		STK_ENDOFDAY		end of trading day
		STK_STARTOFDAY		start of new trading day
		STK_RESET			user quit or seat was revoked, re-init
							shares owned and total value

	The function stk_PriceChange generates stock price changes using
	a combination of random	numbers and some simple calculations
	based on the stock's previous price history.

	Revision History:
	-----------------
	 7/20/90 jmd	preened
	 7/24/90 jmd	added OK to quit? menu
	 9/08/90 jmd	added POSTFEXIT case to aux_BuySell
	11/24/90 jmd	moved to demo, added sfiles
	 1/09/91 pmcm	fixed tickstr array length in aux_CashBox
	 1/18/91 pmcm	added STK_RESET for gameover loss and quit cases
	 1/19/91 bkd	moved lasttime assigment in idle_Stock
	 1/24/91 pmcm	added help button
	 1/24/91 pmcm	changed sed_SetMouseCode to kb_Stuff
	 1/30/91 pmcm	added STKIDLE_HOLDRET STK_IDLE_NOHOLDRET for suppressing
					idle return events while in help
	 1/31/91 pmcm	initialized pricebox data for each demo_Stock call
	 4/03/91 bkd	added "void" to empty argument lists
	 7/20/92 mla	changed sed_SetName() to obj_SetName(); 
					and sed_Find() to obj_Find()
*/

#include "csdemo.h"

#include "kbidle.h"
#include "popdecl.h"
#include "oakalloc.h"		/* 	for omalloc, ofree */

#include <time.h>
#include "cstime.h"
#include "helpdecl.h"		/*	for help_Show() */

/***-------------------------- Local Function Declaration Prototypes -------***/

void 	stock_market(void);
void 	stk_BuySell(int stkno);
int 	stk_PriceChange(sed_type PriceBox);

idle_func (idle_Stock);

aux_func  (aux_CashBox);
aux_func  (aux_PriceBox);
spc_func  (spc_PriceBox);

/***-------------------------- Logicals ------------------------------------***/

/* messages for stock system */

/* new auxiliary messages */
#define STK_CLOCKTICK		AUX_USER(1)
#define STK_PRICECHANGE		AUX_USER(2)		/* indata is stkno */
#define STK_BUYSELLREQ		AUX_USER(3)		/* indata is buysell_struct */
#define STK_BUYSELL			AUX_USER(4)
#define STK_UPDATESTATS		AUX_USER(5)
#define STK_LASTHOUR		AUX_USER(6)
#define STK_ENDOFDAY		AUX_USER(7)
#define STK_STARTOFDAY		AUX_USER(8)
#define STK_RESET			AUX_USER(9)

/* new idle function messages */
#define STKIDLE_NORMAL			(IDLE_LASTMSG + 1)
#define STKIDLE_ENDOFDAY		(IDLE_LASTMSG + 2)
#define STKIDLE_GAMEOVER		(IDLE_LASTMSG + 3)
#define STKIDLE_HOLDRET			(IDLE_LASTMSG + 4)
#define STKIDLE_NOHOLDRET		(IDLE_LASTMSG + 5)

/* other defines */

#define KEY_ENDOFDAY	KEY_USER(0)		/* end of day msg scancode */

#define	INTEREST		1.0002 		  	/* interest rate (per minute) */

#define TICKBUF_LEN		61				/* length of the ticker tape */
#define TICK_TIME		5				/* hsecs time between ticks of ticker */

#define	STK_QUIT		999				/* value of baton when user requests
										 * to quit (set by the "quit" field
										 * and the escape button)
										 */

/***-------------------------- Function Macros -----------------------------***/

/* macros for finding fields in the PriceBox */
#define	pricebox_GetNameField(stkno)		((stkno * 5) + 0)
#define	pricebox_GetCountField(stkno)		((stkno * 5) + 1)
#define	pricebox_GetValueField(stkno)		((stkno * 5) + 2)
#define	pricebox_GetTotalField(stkno)		((stkno * 5) + 3)
#define	pricebox_GetLastField(stkno)		((stkno * 5) + 4)

/***--------------------------- Global Structures for this module ----------***/

/* data structures */

/* current stock information structure */
typedef struct _stock {

	char   *name;			/* name of the stock */

	long	value;			/* current value of the stock */
	long	last;			/* last value paid for the stock */

	long	count;			/* number of shares owned */
	long	total;			/* total value of shares (count * value) */

	int		skew;			/* current price skew of the stock */
	int		trend;			/* current price trend of the stock */

	boolean active;			/* is this an active stock? */

} stock_struct;

/* buy/sell request structure */
typedef struct _buysell {

	int		stkno;			/* the stock number */
	long	count;			/* number of shares to buy (+) or sell (-) */

} buysell_struct;

/* stock price change notification structure */
typedef struct _pricechange {

	int		stkno;			/* the stock number */
	long	price;			/* new share price */

} pricechange_struct;

/* additional data used by the cashbox */
typedef struct _cashbox {

	char	tickbuffer[TICKBUF_LEN + 1];	/* the ticker tape output buffer */
	int		tickpos;						/* current location in the tickbuffer */

	boolean	lasthour;						/* last hour of trading indicator */

} cashbox_struct;

/* additional data used by the pricebox */
typedef struct _pricebox {

	int		volatility;		/* stock activity multiplier */

} pricebox_struct;

/* global array of current stock information */

static stock_struct stk_list[] = {

	/*	name	value	last	count	total	skew	trend	active */

	{ "CHEAPO",	 500L,  500L,  	0L , 	0L,		 0 , 	0, 		TRUE },
	{ "FFC",	1000L, 1000L,  	0L , 	0L,		-1 , 	0, 		TRUE },
	{ "HIFI",	1000L, 1000L,  	0L , 	0L,		 0 , 	0, 		TRUE },
	{ "OAK",	1500L, 1500L,  	0L , 	0L,		 2 , 	0, 		TRUE },
	{ "OHCAL",	1000L, 1000L,  	0L , 	0L,		 0 , 	0, 		TRUE },
	{ "TOSC",	1000L, 1000L,  	0L , 	0L,		 0 , 	0, 		TRUE }
};

/*	initialization data for each time we are called from the demo */
static stock_struct stk_list_init[] = {

	/*	name	value	last	count	total	skew	trend	active */

	{ "CHEAPO",	 500L,  500L,  	0L , 	0L,		 0 , 	0, 		TRUE },
	{ "FFC",	1000L, 1000L,  	0L , 	0L,		-1 , 	0, 		TRUE },
	{ "HIFI",	1000L, 1000L,  	0L , 	0L,		 0 , 	0, 		TRUE },
	{ "OAK",	1500L, 1500L,  	0L , 	0L,		 2 , 	0, 		TRUE },
	{ "OHCAL",	1000L, 1000L,  	0L , 	0L,		 0 , 	0, 		TRUE },
	{ "TOSC",	1000L, 1000L,  	0L , 	0L,		 0 , 	0, 		TRUE }
};

#define STKLIST_COUNT	(sizeof(stk_list) / sizeof(stock_struct))

/* -------------------------------------------------------------------------- */
boolean demo_Stock(void)
/*
	Entry point for the Stock Market Simulator.

	Turns on the keyboard idle function and then calls stock_market().
*/
{
	boolean ret = TRUE;
	/* set time to 12hour format */
	ocountry.time_fmt = TIME_12;

	kb_Idle(idle_Stock);

	/*	clear this 'cuz	 we'll check it to see if the market ran ok later */
	oak_ClearAllocErrorTag();

	stock_market();

	if (oak_GetAllocErrorTag() != 0) {
		/*	allocation error starting stock market */
		oak_ClearAllocErrorTag();
	 	demo_MemOut(demo_Text("memout"));
		ret = FALSE;
	}

	kb_Idle(FNULL);

	return(ret);
}

/* -------------------------------------------------------------------------- */
/* current trading status values */

enum trade_status 	 { TS_NORMAL, TS_ENDOFDAY, TS_STOPPED, TS_GAMEOVER };

int idle_Stock(int msg, unsigned wait)
/*
	The keyboard idle function for the stock exchange simulator.

	The idle function checks the system timer and at the appropriate
	time intervals sends a STK_CLOCKTICK message to the CashBox.

	The idle function also receives the STKIDLE_NORMAL,
	STKIDLE_ENDOFDAY, and STKIDLE_GAMEOVER messages.  These tell the
	idle function to turn on and off.
*/
{
	unsigned 					time;
	static unsigned 			lasttime;
	static enum trade_status 	tstatus;
	static boolean				hold = FALSE;

	oak_notused(wait);

	switch (msg) {

	case IDLE_START:
		lasttime = 0;
		tstatus = TS_NORMAL;
		break;

	case IDLE_STOP:
		break;

	case IDLE_CHECK:
	case IDLE_READ:
		switch(tstatus) {
		case TS_GAMEOVER:
			break;

		case TS_STOPPED:
				/* all trading stopped, continue after a timeout */
				time = hard_Timer();

				if (hard_SubHSecs(lasttime, time) > 500) {
					return(ENTER);
				}
			break;

		case TS_ENDOFDAY:
			if (!hold) {
				/* end of day, stop trading and return ENDOFDAY */
				tstatus = TS_STOPPED;
				lasttime = hard_Timer();
				return(KEY_ENDOFDAY);
			}
			break;

		case TS_NORMAL:
			/* normal trading */
			time = hard_Timer();
			if (hard_SubHSecs(lasttime, time) > TICK_TIME) {
				/* send a TICK message to the ticker tape (the CashBox) */
				sed_DoAux(obj_Find("CashBox"), STK_CLOCKTICK, NULL, NULL);
				lasttime = time;
			}
			break;
		}
		break;

	case STKIDLE_NORMAL:
		/* turn the idle function back on */
		tstatus = TS_NORMAL;
		break;

	case STKIDLE_ENDOFDAY:
		/* end of the day, stop the idle function */
		tstatus = TS_ENDOFDAY;
		break;

	case STKIDLE_GAMEOVER:
		/* game over, stop the idle function */
		tstatus = TS_GAMEOVER;
		break;

	case STKIDLE_HOLDRET:
		/*	do not return any events (other than KEY_NONE) 
		 * 	until further notice
		 */
		hold = TRUE;
		break;

	case STKIDLE_NOHOLDRET:
		/*	ok to return events, clear */
		hold = FALSE;
		break;
	}

	return(KEY_NONE);
}

/* -------------------------------------------------------------------------- */
void stock_market(void)
/*
	Create the CashBox and PriceBox windows for stock trading simulator.
	Actives the system by calling go on the PriceBox.
*/
{

	sed_type	sed, bsed;
	int 		row;
	char 		tickstr[TICKBUF_LEN + 1];
	struct tm	stktime;
	long		cash = 100000L;
	long		average = 1000L;
	long		networth = 100000L;
	long		credit = 100000L;
	int		  	day = 1;

	tickstr[0] = '\0';

	tm_Zero(&stktime);
	stktime.tm_hour = 9;
	stktime.tm_min = 0;

	if ((sed = demo_LoadSed("CashBox", DL_CENTER)) == NULL) {
		return;
	}

	/*	re-point the field variables to our pre-initialized storage.
	 *
	 *	note:	demo_LoadSed calls the library function sfile_LoadSed with
	 *			the SED_ALLOC flag so that field variable storage is 
	 *			allocated with the fields.  
	 *
	 *			We, however, are going to use our storage--even
	 *			though the duplication takes up a smidge more space.
	 *			(Don't worry, sed_Close will free the space allocated by
	 *			demo_LoadSed.)  If we had wanted to use the storage
	 *			demo_LoadSed allocated, then we would call sed_GetVar to get
	 *			a pointer to that store and initialize the store.
	 */

	sed_SetVar(sed, sed_GetNameNo(sed, "cash"), 	(VOID *) &cash);
	sed_SetVar(sed, sed_GetNameNo(sed, "networth"), (VOID *) &networth);
	sed_SetVar(sed, sed_GetNameNo(sed, "credit"), 	(VOID *) &credit);
	sed_SetVar(sed, sed_GetNameNo(sed, "average"), 	(VOID *) &average);
	sed_SetVar(sed, sed_GetNameNo(sed, "day"), 		(VOID *) &day);

	sed_SetVar(sed, sed_GetNameNo(sed, "time"), 	(VOID *) &stktime);
	sed_SetVar(sed, sed_GetNameNo(sed, "ticker"), 	(VOID *) tickstr);

	if ((bsed = sed_GetFieldBob(sed, sed_GetNameNo(sed, "PriceBox"))) != NULL) {
		for (row = 0; row < STKLIST_COUNT; row++) {
			/*	re-point the field variables to our pre-initialized stock
				values.

				note:	this initialization assumes the fields in a row
						for a are arranged in a particular order
			*/
			sed_SetVar(bsed, sed_GetGridField(bsed, row, 0), (VOID *)  stk_list[row].name);
			sed_SetVar(bsed, sed_GetGridField(bsed, row, 1), (VOID *) &stk_list[row].count);
			sed_SetVar(bsed, sed_GetGridField(bsed, row, 2), (VOID *) &stk_list[row].value);
			sed_SetVar(bsed, sed_GetGridField(bsed, row, 3), (VOID *) &stk_list[row].total);
			sed_SetVar(bsed, sed_GetGridField(bsed, row, 4), (VOID *) &stk_list[row].last);
		}
	}

	/*	initialize the stock pricebox data */
	memmove(stk_list, stk_list_init, sizeof(stock_struct) * STKLIST_COUNT);

	sed_Repaint(sed);

	demo_SetButtonData(HELPBTN, (VOID *)sed);
	demo_ShowButton(HELPBTN);

	sed_Go(sed);

	demo_SetButtonData(HELPBTN, NULL);
	demo_HideButton(HELPBTN);

	sed_Close(sed);
}

/* -------------------------------------------------------------------------- */
boolean spc_PriceBox(sed_type sed, int scancode)
/*
	The special function for the PriceBox sed.

	Filters out the Esc key.
	Puts the BuySellBox when the user presses Enter or a numeric key.

	When the idle function returns KEY ENDOFDAY, determines if the game
	is over.  If so puts up the GAMEOVER message and exits.  Otherwise
	puts up the CONTINUE playing message.
*/
{
	int 	ret, key;

	switch (scancode) {

	case HELP_KEY:	/* special help key for certain non-Dos machines */
	case FN1:
		/*	let idle function keep any return values to itself 
		 *	while we are in help
		 */
		idle_Stock(STKIDLE_HOLDRET, 0);
		oak_ClearAllocErrorTag();
		help_Show(sed_GetLabel(sed), sed_GetFieldNo(sed) + 1);
		if (oak_GetAllocErrorTag() != 0) {
			oak_ClearAllocErrorTag();
			demo_MemOut(demo_Text("helpmemout"));
		}
		idle_Stock(STKIDLE_NOHOLDRET, 0);
		return(TRUE);

	case ESC:
		/*	pre-empt it */
		return(TRUE);

	case ENTER:
	case MOU_CLICK:
		if (!sed_IsFieldName(sed, sed_GetFieldNo(sed), "quit")) {
			/*	the BuySell box has a greedy handler,
			 *	the buttons are of no use
			 */
			demo_HideButton(HELPBTN);
			stk_BuySell(sed_GetGridRow(sed, sed_GetFieldNo(sed)));
			demo_ShowButton(HELPBTN);
			return(TRUE);
		}
		break;

	case KEY_ENDOFDAY:
		/* this scancode is generated by the keyboard idle function */
		/* send ENDOFDAY message to PriceBox and the CashBox */
		ret = sed_DoAux(sed, STK_ENDOFDAY, NULL, NULL);

		if (ret != 0) {
			ret = sed_DoAux(obj_Find("CashBox"), STK_ENDOFDAY, NULL, NULL);
		}

		if (ret == 0) {
			/* you lose, tell the idle function to stop */
			idle_Stock(STKIDLE_GAMEOVER, 0);
			hard_Speaker(2500, 50);

			demo_HideButton(HELPBTN);

			demo_LoadSed("StkGameOver", DL_PAINT | DL_CENTER | DL_GO | DL_CLOSE | DL_MEMOUT);

			demo_ShowButton(HELPBTN);

			/* 	reset shares owned and total stock value */
			sed_DoAux(sed, STK_RESET, NULL, NULL);

			/* set baton to SED_ABORT to turn off Quit message in aux func */
			sed_SetBaton(sed, SED_ABORT);
			sed_ToggleExit(sed);
			return(TRUE);
		}
		else {
			demo_HideButton(HELPBTN);

			demo_LoadSed("StkContinue", DL_PAINT | DL_CENTER | DL_GO | DL_CLOSE | DL_MEMOUT);

			demo_ShowButton(HELPBTN);

			/* resume trading, tell the idle function to start up again */
			idle_Stock(STKIDLE_NORMAL, 0);
			sed_DoAux(obj_Find("CashBox"), STK_STARTOFDAY, NULL, NULL);
		}
		break;

	default:
		key = ascii(scancode);
		if (isdigit(key) || key == '-') {
			if (!sed_IsFieldName(sed, sed_GetFieldNo(sed), "quit")) {
				kb_Stuff(scancode);
				stk_BuySell(sed_GetGridRow(sed, sed_GetFieldNo(sed)));
				return(TRUE);
			}
		}
		break;
	}

	return(FALSE);
}

/* -------------------------------------------------------------------------- */
int aux_CashBox(sed_type sed, int msg, VOID *indata, VOID *outdata)
/*
	Auxiliary function for the CashBox.

	Manages the ticker tape, time clock, and current financial data.
*/
{
	cashbox_struct *cashboxp;
	char 			tickstr[TICKBUF_LEN + 2];
	int				fno, stkno;
	long			total, cash, credit, purchase, *cashp;
	buysell_struct *buysellp;
	struct tm 	   *timep;
	int			   *dayp;

	if (msg != AUX_START) {
		cashboxp = (cashbox_struct *) sed_GetData(sed);
	}

	switch(msg) {
	case AUX_START:
		/*	allocate a cash box data struct, attach it to the sed data
		 *	pointer
		 */
		if ((cashboxp = (cashbox_struct *) omalloc(OA_NOTAG, sizeof(cashbox_struct))) != NULL) {
			sed_SetData(sed, (VOID *) cashboxp);
			cashboxp->tickpos = -1;
			cashboxp->lasthour = FALSE;
		}
		else {
			return(0);
		}
		break;

	case AUX_END:
		/* free the cashbox data struct */
		ofree(OA_NOTAG, (cashbox_struct *) sed_GetData(sed));
		sed_SetData(sed, NULL);
		break;

	case STK_CLOCKTICK:
		/* 	(sent by Idle function)
		 * 	update the time of day clock
		 * 	update the ticker tape
		 * 	if the ticker buffer is empty, send the tick to the PriceBox
		 * 	so that it can generate a new price change.
		 */
		/* update time of day clock */
		timep = (struct tm *) sed_GetNameVar(sed, "time");
		tm_Copy(timep, tm_AddSecs(timep, 15L));

		if (timep->tm_hour >= 16) {
			/* it's after 4:00 pm */
			hard_Speaker(2200, 30);

			/* tell the idle function that the day is ended */
			idle_Stock(STKIDLE_ENDOFDAY, 0);
		}
		else if (timep->tm_hour >= 15 && !cashboxp->lasthour) {
			/* it's after 3:00 pm */
			cashboxp->lasthour = TRUE;

			/* send LASTHOUR message to PriceBox */
			sed_DoAux(obj_Find("PriceBox"), STK_LASTHOUR, NULL, NULL);
		}

		/* if it's an exact minute update the clock field (and charge interest) */
		if (timep->tm_sec == 0) {
			sed_RepaintField(sed, sed_GetNameNo(sed, "time"));

			/* test for interest payment */
			cashp = (long *) sed_GetNameVar(sed, "cash");
			if (*cashp < 0L) {
				*cashp = (long) ((float) *cashp * INTEREST);
				sed_RepaintField(sed, sed_GetNameNo(sed, "cash"));

				/* update the networth indicator */
				sed_DoAux(sed, STK_UPDATESTATS, NULL, NULL);
			}

		}

		/* check the ticker buffer */
		if (cashboxp->tickpos < 0) {
			/* 	empty, send tick to PriceBox */
			sed_DoAux(obj_Find("PriceBox"), STK_CLOCKTICK, indata, outdata);
		}
		else {
			/* 	not empty, update ticker tape */
			fno = sed_GetNameNo(sed, "ticker");
			tickstr[0] = cashboxp->tickbuffer[cashboxp->tickpos];
			strcpy(tickstr + 1, sed_GetRecord(sed, fno));
			sed_SetRecord(sed, tickstr, fno);
			sed_UpdateField(sed, fno);
			cashboxp->tickpos--;
		}
		break;

	case STK_PRICECHANGE:
		/* 	Put new entry into the ticker buffer,
		 * 	send ourselves an UPDATESTATS message to update the networth indicator
		 */

		stkno = *((int *) indata);

		/* put the new data into the ticker buffer */
		if (stk_list[stkno].active) {
			sprintf(cashboxp->tickbuffer, "%s %.2f  ", stk_list[stkno].name, (float) (stk_list[stkno].value/100.0));
		}
		else {
			/* the stock is bankrupt */
			sprintf(cashboxp->tickbuffer, "%s %s  ", stk_list[stkno].name, demo_Text("bankrupt"));
		}
		cashboxp->tickpos = strlen(cashboxp->tickbuffer) - 1;

		sed_DoAux(sed, STK_UPDATESTATS, NULL, NULL);
		break;

	case STK_UPDATESTATS:
		/* update the stock average and networth indicators */

		/* compute new average */
		for (stkno = 0, total = 0L; stkno < STKLIST_COUNT; stkno++) {
			total += stk_list[stkno].value;
		}

		fno = sed_GetNameNo(sed, "average");
		*((long *) sed_GetVar(sed, fno)) = total / STKLIST_COUNT;
		sed_RepaintField(sed, fno);

		/* compute new networth */
		/* total = cash + value of owned stocks */
		total = *((long *) sed_GetNameVar(sed, "cash"));

		for (stkno = 0; stkno < STKLIST_COUNT; stkno++) {
			total += stk_list[stkno].total;
		}

		fno = sed_GetNameNo(sed, "networth");
		*((long *) sed_GetVar(sed, fno)) = total;
		sed_RepaintField(sed, fno);
		break;

	case STK_BUYSELLREQ:
		/* 	Test if a transaction can be made:
		 * 	if so, send a BUYSELL message to the PriceBox
		 *  and update the cash indicator
		 *  send ourselves an UPDATESTATS message to update the networth
		 *	indicator; otherwise fail (return 0)
		 */
		buysellp = (buysell_struct *) indata;

		/* can we afford it? */
		purchase = (buysellp->count * stk_list[buysellp->stkno].value);
		cash = *((long *) sed_GetNameVar(sed, "cash")) - purchase;
		credit = *((long *) sed_GetNameVar(sed, "credit"));

		/* if we're buying something, see if we can afford it */
		if (!stk_list[buysellp->stkno].active || (purchase > 0L && cash < -credit)) {
			return(0);
		}

		/* update the cash indicator */
		*((long *) sed_GetNameVar(sed, "cash")) = cash;

		sed_RepaintField(sed, sed_GetNameNo(sed, "cash"));

		/* now send a buysell message to the PriceBox */
		sed_DoAux(obj_Find("PriceBox"), STK_BUYSELL, (VOID *)buysellp, NULL);

		/* send an UPDATESTATS message to ourselves */
		sed_DoAux(sed, STK_UPDATESTATS, NULL, NULL);
		break;

	case STK_STARTOFDAY:
		/* 	start of a new day, reset the clock increment the day counter */
		timep = (struct tm *) sed_GetNameVar(sed, "time");
		timep->tm_hour = 9;
		cashboxp->lasthour = FALSE;
		sed_RepaintField(sed, sed_GetNameNo(sed, "time"));

		/* update the day indicator */
		dayp = (int *) sed_GetNameVar(sed, "day");
		(*dayp)++;
		sed_RepaintField(sed, sed_GetNameNo(sed, "day"));
		break;

	case STK_ENDOFDAY:
		/* 	if our total value is negative we lose (return 0) */
		if (*((long *) sed_GetNameVar(sed, "networth")) < 0L) {
			return(0);
		}

		/* 	update our credit limit */
		if (*((long *) sed_GetNameVar(sed, "cash")) < 0L) {
			*((long *) sed_GetNameVar(sed, "credit")) = (long)
			((float)(*((long *) sed_GetNameVar(sed, "credit"))) * ((float) 0.9));
		}
		else {
			*((long *) sed_GetNameVar(sed, "credit")) = (long) 
			((float) (*((long *) sed_GetNameVar(sed, "credit"))) * ((float) 1.1));
		}

		sed_RepaintField(sed, sed_GetNameNo(sed, "credit"));
		break;

	case STK_BUYSELL:
	case STK_LASTHOUR:
		/* ignore */
		break;
	}

	return(1);
}

/* -------------------------------------------------------------------------- */
int aux_PriceBox(sed_type sed, int msg, VOID *indata, VOID *outdata)
/*
	Auxiliary function for PriceBox object.

	Generate price changes and handles updating the list of current
	stock information.
*/
{
	pricebox_struct    *pricep;
	int 				row;
	int 				stkno;
	int					ret;
	buysell_struct 	   *buysellp;
	buysell_struct 		buysell;

	oak_notused(outdata);

	if (msg != AUX_START) {
		pricep = (pricebox_struct *) sed_GetData(sed);
	}

	switch(msg) {
	case AUX_START:
		/* 	allocate a pricebox data struct, attach it to the sed data
	     *	pointer
		 */
		if ((pricep = (pricebox_struct *) omalloc(OA_NOTAG, sizeof(pricebox_struct))) != NULL) {
			sed_SetData(sed, (VOID *) pricep);
			pricep->volatility = 1;
		}
		else {
			return(0);
		}
		break;

	case AUX_END:
		/* free the pricebox data struct */
		ofree(OA_NOTAG, (pricebox_struct *) sed_GetData(sed));
		sed_SetData(sed, NULL);
		break;

	case SED_PRESEXIT:
		/* ask if it's OK to quit? */
		if (sed_GetBaton(sed) != SED_ABORT) {
			/*	don't put up quit message if we are leaving to
			 *	move to another window
			 */
			if (sed_GetBaton(sed) == STK_QUIT) {
				demo_HideButton(HELPBTN);
				oak_ClearAllocErrorTag();

				ret = demo_Query(demo_Text("stock_quit"));

				if (oak_GetAllocErrorTag() != 0) {
					/* just quit if we cannot query */
					oak_ClearAllocErrorTag();
					ret = 1;			
				}

				demo_ShowButton(HELPBTN);

				if (ret != 1) {
					/*	escape or no */
					return(0);
				}
				else {
					/* yes, quit */
					sed_SetBaton(sed, SED_ABORT);

					/*	send ourself the STK_RESET message */
					sed_DoAux(sed, STK_RESET, NULL, NULL);
				}
			}
		}
		break;

	case STK_CLOCKTICK:
		/* 	(sent by CashBox) generate a price change,
		 * 	send PRICECHANGE messages to ourselves, the CashBox and
		 *	the BuySellBox
		 */

		/* generate a stock price change */
		stkno = stk_PriceChange(sed);

		/* Send Price change messages to the various boxes */
		sed_DoAux(sed, STK_PRICECHANGE, (VOID *)&stkno, NULL);
		sed_DoAux(obj_Find("CashBox"), STK_PRICECHANGE, (VOID *)&stkno, NULL);
		sed_DoAux(obj_Find("BuySellBox"), STK_PRICECHANGE, (VOID *)&stkno, NULL);
		break;

	case STK_PRICECHANGE:
		/* update the stock total value, Refresh the stock price list */
		stkno = *((int *) indata);

		stk_list[stkno].total = stk_list[stkno].value * stk_list[stkno].count;

		if (!stk_list[stkno].active) {
			/* the stock is kaput! fix the PriceBox */

			/* first protect and mark the field */
			sed_ProtectField(sed, pricebox_GetNameField(stkno));
			sed_MarkField(sed, pricebox_GetNameField(stkno), 0x14, 0x14);
			sed_UpdateField(sed, pricebox_GetNameField(stkno));

			/* if the current field becomes protected, go to another field */
			if (sed_IsProtectedField(sed, sed_GetFieldNo(sed))) {
				sed_GotoFirstField(sed);
			}

			/* update number of shares owned */
			sed_RepaintField(sed, pricebox_GetCountField(stkno));
		}
		else if (sed_IsProtectedField(sed, pricebox_GetNameField(stkno))) {
			/* share is active again */
			sed_UnProtectField(sed, pricebox_GetNameField(stkno));
			sed_UnMarkField(sed, pricebox_GetNameField(stkno));
			sed_RepaintField(sed, pricebox_GetNameField(stkno));
		}

		sed_RepaintField(sed, pricebox_GetValueField(stkno));
		sed_RepaintField(sed, pricebox_GetTotalField(stkno));
		break;

	case STK_BUYSELLREQ:
		/* ignore */
		break;

	case STK_BUYSELL:
		/* change the number of shares owned */
		buysellp = (buysell_struct *) indata;
		stkno = buysellp->stkno;

		/* update number of shares owned */
		stk_list[stkno].count += buysellp->count;
		sed_RepaintField(sed, pricebox_GetCountField(stkno));

		/* update total price for share */
		stk_list[stkno].total = stk_list[stkno].value * stk_list[stkno].count;
		sed_RepaintField(sed, pricebox_GetTotalField(stkno));

		/* set the last price paid */
		stk_list[stkno].last = stk_list[stkno].value;
		sed_RepaintField(sed, pricebox_GetLastField(stkno));
		break;

	case STK_UPDATESTATS:
		/* ignore */
		break;

	case STK_LASTHOUR:
		/* last hour of trading warning: raise volatility level */
		/* also sent during panics */
		pricep->volatility++;
		break;

	case STK_ENDOFDAY:
		/* end of trading day: reset volatility level, settle short shares */
		pricep->volatility = 1;

		for (stkno = 0; stkno < STKLIST_COUNT; stkno++) {
			if (stk_list[stkno].count < 0L && stk_list[stkno].active) {
				/* send a buy request for the short stock to the CashBox */
				buysell.stkno = stkno;
				buysell.count = -stk_list[stkno].count;

				if (sed_DoAux(obj_Find("CashBox"), STK_BUYSELLREQ, (VOID *)&buysell, NULL) == 0) {
					/* unable to settle shares, you lose */
					return(0);
				}
			}
		}
		break;

	case STK_RESET:
		/* 	sent by our special function in response to the loss of seat
		 *	on exchange or, by ourself in response to quit.
		 */
		for (row = 0; row < STKLIST_COUNT; row++) {
			stk_list[row].count	= 0L;
			stk_list[row].total	= 0L;
		}
		break;
	}

	return(1);
}

/* -------------------------------------------------------------------------- */
void stk_BuySell(int stkno)
/*
	Creates and activates the BuySell dialogue box for the stock 'stkno'.
*/
{
	sed_type sed;
	long 	 count = 0L;

	if ((sed = demo_LoadSed("BuySellBox", DL_CENTER | DL_MEMOUT)) == NULL) {
	 	return;
	}

	/* attach data to sed */
	sed_SetNameVarValue(sed, "name",  (VOID *) stk_list[stkno].name);
	sed_SetNameVarValue(sed, "count", (VOID *) &count);
	sed_SetNameVarValue(sed, "stkno", (VOID *) &stkno);

	/* in this case we want the field to point to the original data */
	sed_SetNameVar(sed, "price", (VOID *) &stk_list[stkno].value);

	obj_SetName(sed, "BuySellBox");

	sed_Repaint(sed);
	sed_Go(sed);
	sed_Close(sed);
}

/* -------------------------------------------------------------------------- */
boolean spc_BuySellBox(sed_type sed, int scancode)
/*
	The special function for the BuySellBox sed.

	Handles the error message for the box,
	and intercepts the KEY_ENDOFDAY scancode (which removes the box).
*/
{
	switch (scancode) {
	
	case HELP_KEY:	/* special help key for certain non-Dos machines */
	case FN1:
		/*	let idle function keep any return values to itself 
		 *	while we are in help
		 */
		idle_Stock(STKIDLE_HOLDRET, 0);
		oak_ClearAllocErrorTag();
		help_Show(sed_GetLabel(sed), sed_GetFieldNo(sed) + 1);
		if (oak_GetAllocErrorTag() != 0) { 
			oak_ClearAllocErrorTag();
	 		demo_MemOut(demo_Text("helpmemout"));
		}
		idle_Stock(STKIDLE_NOHOLDRET, 0);
		return(TRUE);

	case KEY_INVALID:
		/* The validation in fexit failed: tell the user */
		tone();
		sed_BorderPrompt(sed, demo_Text("no_afford_msg"));

		/* wait for a keystroke
		 * (use kb_Check to make sure the idle func gets called)
		 */
		while (!kb_Check()) {
			;
		}

		sed_BorderPrompt(sed, NULL);
		return(TRUE);

	case KEY_ENDOFDAY:
		/* this scancode is generated by the keyboard idle function */
		/* remove the popup, stuff ENDOFDAY into mousecode for PriceBox */
		sed_SetBaton(sed, 0);
		sed_ToggleExit(sed);
		kb_Stuff(KEY_ENDOFDAY);
		return(TRUE);

	default:
		break;
	}

	return(FALSE);
}

/* -------------------------------------------------------------------------- */
int aux_BuySellBox(sed_type sed, int msg, VOID *indata, VOID *outdata)
/*
	Auxiliary function for BuySell object.

	Validates stock purchase requests
	and updates the displayed stock price when it changes.
*/
{
	buysell_struct buysell;
	int 	stkno;

	oak_notused(outdata);

	switch(msg) {

	case SED_POSTFEXIT:
		/* move record data into field variable */
		sed_DoFieldSexit(sed, sed_GetFieldNo(sed));
		break;

	case SED_PRESEXIT:
		/* Send a BUYSELLREQ msg to the CashBox
		 * fail if the msg fails
		 */

		if (sed_GetBaton(sed) != 0) {
			/* first, send a buy request to the CashBox */
			buysell.stkno = *((int *) sed_GetNameVar(sed, "stkno"));
			buysell.count = *((long *) sed_GetNameVar(sed, "count"));

			if (sed_DoAux(obj_Find("CashBox"), STK_BUYSELLREQ, (VOID *)&buysell, NULL) == 0) {
			 	kb_Stuff(KEY_INVALID);
				return(0);
			}
		}
		break;

	case STK_CLOCKTICK:
		/* ignore */
		break;

	case STK_PRICECHANGE:
		/* update the stock price (if applicable) */
		stkno = *((int *) indata);
		if (stkno == *((int *) sed_GetNameVar(sed, "stkno"))) {
			sed_RepaintField(sed, sed_GetNameNo(sed, "price"));
		}
		break;

	case STK_BUYSELLREQ:
	case STK_BUYSELL:
	case STK_UPDATESTATS:
	case STK_LASTHOUR:
	case STK_ENDOFDAY:
		/* ignore */
		break;
	}

	return(1);
}
/* -------------------------------------------------------------------------- */

/* matrix of price changes (in 1/10 of a percent) */
static int stk_matrix[] =
{ -100,-80,-10,-8,-6,-5,-4,-3,-2,-1,0,1,2,3,4,6,8,10,15,80,100 };

#define MATRIX_LEN	(sizeof(stk_matrix) / sizeof(int))

int stk_PriceChange(sed_type PriceBox)
/*
	Called by the PriceBox.
	Generates a price change and updates the information
	in the stock struct list.
	Returns the number of the changed stock.
*/
{
	int		change, stkno, i;
	float	pct;
	pricebox_struct *pricep;

	pricep = (pricebox_struct *) sed_GetData(PriceBox);

	/* choose a stock (make sure it's active) */
	while(1) {
		stkno = rand() % STKLIST_COUNT;
		if (stk_list[stkno].active) {
			break;
		}
		else {
			/* inactive stock, test if it comes back to life */
			if (rand() % 100 < 1) {
				stk_list[stkno].value = (long) (rand() % 1000) + 25L;
				stk_list[stkno].active = TRUE;
				break;
			}
		}
	}

	/* set its value */
	/* compute a change value using the stock matrix */
	change = (rand() % MATRIX_LEN) + stk_list[stkno].skew;
	change = (change < 0) ? 0 : change;
	change = (change >= MATRIX_LEN) ? MATRIX_LEN - 1 : change;
	change = stk_matrix[change] * pricep->volatility;

	/* compute the trend */
	if (change > 0 && stk_list[stkno].value > 100L) {
		if (stk_list[stkno].value > 10000L) {
			stk_list[stkno].trend += pricep->volatility * 4;
		}
		else if (stk_list[stkno].value > 5000L) {
			stk_list[stkno].trend += pricep->volatility * 2;
		}
		else {
			stk_list[stkno].trend += pricep->volatility;
		}
	}
	else {
		stk_list[stkno].trend--;
		stk_list[stkno].trend = (stk_list[stkno].trend < 0) ? 0 : stk_list[stkno].trend;
	}

	/* test for "market correction" */
	if (rand() % 1000 < stk_list[stkno].trend) {
		change = (rand() % 400) - 450;

		tone();

		/* increase everyone's chance of correction */
		for (i = 0; i < STKLIST_COUNT; i++) {
			stk_list[i].trend += (rand() % 10);
		}

		stk_list[stkno].skew = 0;
		stk_list[stkno].trend = 0;
	}

	/* test for "rally" */
	if (rand() % 1000 < 1) {
		hard_Speaker(1500, 10);
		hard_Speaker(500, 10);

		/* zero everyone's chance of correction, up their skew */
		for (i = 0; i < STKLIST_COUNT; i++) {
			stk_list[i].trend = 0;
			stk_list[i].skew = 3;
		}
	}


	/* test for "panic" */
	if (rand() % 2200 < 1) {

		hard_Speaker(1700, 100);
		hard_Speaker(750, 10);
		hard_Speaker(1700, 10);
		hard_Speaker(750, 100);

		/* shrink everyone's skew */
		for (i = 0; i < STKLIST_COUNT; i++) {
			stk_list[i].skew = -3;
		}

		/* send LASTHOUR message to increase volatility */
		sed_DoAux(obj_Find("PriceBox"), STK_LASTHOUR, NULL, NULL);
	}

	/* change the skew */
	if (rand() % 10 < 1) {
		stk_list[stkno].skew = (rand() % 6) - 2;
	}

	/* compute the new value */
	pct = (float) 1.0 + ((float) change / (float) 1000.0);

	stk_list[stkno].value =	(long) ((float) stk_list[stkno].value * pct);

	if (stk_list[stkno].value < 25L) {
		stk_list[stkno].skew = 4;
		if (rand() % 50 < 1 || stk_list[stkno].value < 1L) {
			/* company fails */
			stk_list[stkno].value = 0L;
			stk_list[stkno].active = FALSE;
			stk_list[stkno].total = 0L;

			/* clear number of shares owned (unless it's short) */
			if (stk_list[stkno].count > 0L) {
				stk_list[stkno].count = 0L;
			}
		}
	}

	return(stkno);
}
/* -------------------------------------------------------------------------- */
